
#include "huffmantree.h"
using namespace std;

namespace Huffman
{

    TreeNode::TreeNode(char c)
    {
        children[0] = NULL;
        children[1] = NULL;
        count = 0;
        character = c;
    }

    TreeNode::~TreeNode()
    {
        if(children[0] != NULL)
            delete children[0];
        if(children[1] != NULL)
            delete children[1];
    }

    int TreeNode::Compare(TreeNode* a, TreeNode* b) // for the sorting
    {
        if(a->count > b->count)
            return 1;
        if(a->count < b->count)
            return -1;
        return 0;
    }
    
    
    
    
    
    HuffmanTree::HuffmanTree()
    {
        root = NULL;
    }
    
    HuffmanTree::~HuffmanTree()
    {
        if(root != NULL)
            delete root;
    }

    int HuffmanTree::ConstructFromStream(istream& is) // Construct the tree, using the Huffman-Algorithm
    {
        if(root != NULL)
            delete root;
        root = NULL;
        
        vector<TreeNode*> table(256, NULL);
        for(int i = table.size()-1; i >= 0; --i)
            table[i] = new TreeNode((char)i);
        
        is.peek();
        while(is.good())
        {
            unsigned char c = is.get();
            table[(int)c]->count++;
            is.peek(); // update the eof bit
        }
        SelectionSort(table, TreeNode::Compare);
        
        for(int i = table.size()-1; i > 0; --i)
        {
            if(table[i]->count <= 0)
            {
                delete table[i];
                table[i] = NULL;
                continue;
            }

            TreeNode* t = new TreeNode();
            t->count = table[i-1]->count + table[i]->count;
            t->children[0] = table[i-1];
            t->children[1] = table[i];
            table[i] = NULL;
            table[i-1] = t;
            
            InsertionSortOnce(table, i-1, TreeNode::Compare);
        }
        
        root = table[0];
        return root->count;
    }
    
    
    void HuffmanTree::SaveToStream(ostream& os)
    {
        if(root == NULL)
            return;

        bitostream bos(os);
        
        stack<TreeNode*> s;
        vector<TreeNode*> leaves;
        s.push(root);
        
        // first, we only store the structure of the tree, followed by
        // the contents of the leaves from left to right (saves a lot of bytes!)
        while(!s.empty())
        {
            TreeNode* node = s.top();
            s.pop();
        
            if(node->children[0] == NULL) // if node is a leaf
            {
                bos << 0;
                leaves.push_back(node);
            }
            else
            {
                bos << 1;
                s.push(node->children[1]);
                s.push(node->children[0]);
            }
        }
        
        bos.flush();
        for(vector<TreeNode*>::iterator i = leaves.begin(); i != leaves.end(); i++)
            os << (*i)->character;
    }
    
    
    void HuffmanTree::LoadFromStream(istream& is)
    {
        if(root != NULL)
            delete root;
        root = NULL;
        
        bitistream bis(is);
        vector<TreeNode*> leaves;
        
        stack<TreeNode**> s;
        s.push(&root);

        while(!s.empty())
        {
            TreeNode** top = s.top();
            s.pop();
            *top = new TreeNode();
            
            bool haschildren;
            bis >> haschildren;
            if(haschildren)
            {
                s.push(&(*top)->children[1]);
                s.push(&(*top)->children[0]);
            }
            else
            {
                leaves.push_back(*top);
            }
        }
        
        for(vector<TreeNode*>::iterator i = leaves.begin(); i != leaves.end(); i++)
            (*i)->character = is.get();
    }

    
    
    void HuffmanTree::CreateSymbolTable(TreeNode* node, vector<bool> path, vector<vector<bool> > *symboltable)
    {
        if(node->children[0] == NULL)
        {
            symboltable->operator[](node->character) = path;
        }
        else
        {
            vector<bool> path0 = path;
            path0.push_back(0);
            CreateSymbolTable(node->children[0], path0, symboltable);
            vector<bool> path1 = path;
            path1.push_back(1);
            CreateSymbolTable(node->children[1], path1, symboltable);
        }
    }
    void HuffmanTree::Encode(istream& is, ostream& os)
    {
        if(root == NULL)
            return;

        // create the symbol table
        vector<vector<bool> > symboltable(256, vector<bool>());
        CreateSymbolTable(root, vector<bool>(0), &symboltable);

        // encode the stream
        bitostream bos(os);
        for(is.peek(); !is.eof(); is.peek())
        {
            char c = is.get();
            for(vector<bool>::iterator i = symboltable[c].begin(); i != symboltable[c].end(); i++)
                bos << *i;
        }
        bos.flush();
    }
    
    
    void HuffmanTree::Decode(istream& is, int numberofbytes, ostream& os)
    {
        if(root == NULL || root->children[0] == NULL)
            return;
        
        TreeNode* run = root;
        bitistream bis(is);
        bool direction;
        
        while(numberofbytes > 0)
        {
            bis >> direction;
            run = run->children[direction];
            if(run->children[0] == NULL)
            {
                os.put(run->character);
                run = root;
                numberofbytes--;
            }
        }
    }
    
}
